import arrow
from tinydb import Query, TinyDB
from tinydb.operations import set

from .config import config


async def exist_in_cache(db: TinyDB, image_md5: str, mode: str) -> dict:
    cache_result = db.search((Query().image_md5 == image_md5) & (Query().mode == mode))
    if cache_result:
        db.update(
            set("update_at", arrow.now().for_json()), Query().image_md5 == image_md5
        )
        return cache_result[-1]
    else:
        return {}


async def clear_expired_cache(db: TinyDB) -> None:
    expired_date = arrow.now().shift(days=-config.cache_expire).for_json()
    db.remove(Query().update_at < expired_date)



from PicImageSearch import AsyncAscii2D, NetWork
from PicImageSearch.Utils import Ascii2DResponse

from .utils import handle_img, shorten_pixiv_url


async def ascii2d_search(url: str, proxy: str, hide_img: bool) -> list[str]:
    async with NetWork(proxy=proxy) as client:
        ascii2d_color = AsyncAscii2D(client=client)
        ascii2d_bovw = AsyncAscii2D(bovw=True, client=client)
        color_res = await ascii2d_color.search(url)
        bovw_res = await ascii2d_bovw.search(url)

        async def get_final_res(res: Ascii2DResponse) -> str:
            if not res.raw[0].url:
                res.raw[0] = res.raw[1]
            thumbnail = await handle_img(res.raw[0].thumbnail, proxy, hide_img)
            res_list = [
                f"{thumbnail}",
                f"{res.raw[0].title}" if res.raw[0].title else "",
                f"Author：{res.raw[0].authors}" if res.raw[0].authors else "",
                f"{await shorten_pixiv_url(res.raw[0].url)}" if res.raw[0].url else "",
            ]
            return "\n".join([i for i in res_list if i != ""])

        color_final_res = await get_final_res(color_res)
        bovw_final_res = await get_final_res(bovw_res)
        if color_final_res == bovw_final_res:
            return [f"Ascii2D 色合検索与特徴検索結果完全一致\n{color_final_res}"]

        return [
            f"Ascii2D 色合検索結果\n{color_final_res}",
            f"Ascii2D 特徴検索結果\n{bovw_final_res}",
        ]



from PicImageSearch import AsyncSauceNAO, NetWork

from .ascii2d import ascii2d_search
from .config import config
from .utils import get_source, handle_img, shorten_pixiv_url


async def saucenao_search(url: str, mode: str, proxy: str, hide_img: bool) -> list[str]:
    saucenao_db = {"all": 999, "pixiv": 5, "danbooru": 9, "anime": 21, "doujin": 38}
    async with NetWork(proxy=proxy) as client:
        saucenao = AsyncSauceNAO(
            client=client, api_key=config.saucenao_api_key, db=saucenao_db[mode]
        )
        res = await saucenao.search(url)
        final_res = []
        thumbnail = await handle_img(res.raw[0].thumbnail, proxy, hide_img)
        source = await shorten_pixiv_url(await get_source(res.raw[0].url, proxy))
        if res is not None:
            res_list = [
                f"SauceNAO（{res.raw[0].similarity}%）",
                f"{thumbnail}",
                f"{res.raw[0].origin['data'].get('jp_name')}"
                if mode == "doujin"
                else f"{res.raw[0].title}",
                f"Author：{res.raw[0].author}" if res.raw[0].author else "",
                f"{await shorten_pixiv_url(res.raw[0].url)}",
                f"Source：{source}" if source else "",
            ]
            final_res.append("\n".join([i for i in res_list if i != ""]))
            if (
                config.use_ascii2d_when_low_acc
                and res.raw[0].similarity < config.saucenao_low_acc
            ):
                final_res.append(f"相似度 {res.raw[0].similarity}% 过低，自动使用 Ascii2D 进行搜索")
                final_res.extend(await ascii2d_search(url, proxy, hide_img))
        else:
            final_res.append("SauceNAO 暂时无法使用，自动使用 Ascii2D 进行搜索")
            final_res.extend(await ascii2d_search(url, proxy, hide_img))
        return final_res



import re

import arrow
from nonebot.adapters.onebot.v11 import (
    Bot,
    GroupMessageEvent,
    Message,
    MessageEvent,
    PrivateMessageEvent,
)
from nonebot.matcher import Matcher
from nonebot.params import CommandArg
from nonebot.plugin.on import on_command, on_message
from nonebot.rule import Rule
from tinydb import JSONStorage, Query, TinyDB
from tinydb.middlewares import CachingMiddleware

from .ascii2d import ascii2d_search
from .cache import clear_expired_cache, exist_in_cache
from .config import config
from .saucenao import saucenao_search

if config.proxy:
    PROXY = config.proxy
else:
    PROXY = None


async def _to_me(bot: Bot, event: MessageEvent) -> bool:
    msgs = event.message
    at_me = bool([i for i in msgs if i.type == "at" and i.data["qq"] == bot.self_id])
    if event.reply:
        has_image = bool([i for i in event.reply.message if i.type == "image"])
    else:
        has_image = bool([i for i in msgs if i.type == "image"])
    return has_image and (event.to_me or at_me)


async def _not_to_me(event: MessageEvent) -> bool:
    return not event.to_me and isinstance(event, GroupMessageEvent)


IMAGE_SEARCH = on_message(rule=Rule(_to_me), priority=5)
IMAGE_SEARCH_MODE = on_command("搜图", rule=Rule(_not_to_me), priority=5)


@IMAGE_SEARCH_MODE.handle()
async def handle_first_receive(matcher: Matcher, args: Message = CommandArg()) -> None:
    if [i for i in args if i.type == "image"]:
        matcher.set_arg("IMAGES", args)


async def image_search(
    url: str, mode: str, purge: bool, proxy: str, hide_img: bool = config.hide_img
) -> list[str]:
    db = TinyDB(
        "cache.json",
        storage=CachingMiddleware(JSONStorage),
        encoding="utf-8",
        sort_keys=True,
        indent=4,
        ensure_ascii=False,
    )
    image_md5 = re.search("[A-F0-9]{32}", url)[0]
    result = await exist_in_cache(db, image_md5, mode)
    cached = bool(result)
    if purge or not cached:
        result = {}
        if mode == "a2d":
            result["ascii2d"] = await ascii2d_search(url, proxy, hide_img)
        else:
            result["saucenao"] = await saucenao_search(url, mode, proxy, hide_img)
        result["mode"] = mode
        result["image_md5"] = image_md5
        result["update_at"] = arrow.now().for_json()
        db.upsert(result, (Query().image_md5 == image_md5) & (Query().mode == mode))
        db.insert(result)
    await clear_expired_cache(db)
    db.close()
    if mode == "a2d":
        final_res = result["ascii2d"]
    else:
        final_res = result["saucenao"]
    if cached and not purge:
        final_res = [f"[缓存] {i}" for i in final_res]
    return final_res


async def get_image_urls(msg: Message) -> list[str]:
    return [i.data["url"] for i in msg if i.type == "image" and i.data.get("url")]


async def get_args(msg: Message) -> (str, bool):
    mode = "all"
    plain_text = msg.extract_plain_text()
    args = ["a2d", "pixiv", "danbooru", "doujin", "anime"]
    if plain_text:
        for i in args:
            if f"--{i}" in plain_text:
                mode = i
    purge = "--purge" in plain_text
    return mode, purge


@IMAGE_SEARCH.handle()
@IMAGE_SEARCH_MODE.got("IMAGES", prompt="请发送图片及搜索类型（可选）")
async def handle_image_search(bot: Bot, event: MessageEvent) -> None:
    message = event.reply.message if event.reply else event.message
    image_urls = await get_image_urls(message)
    if not image_urls:
        await IMAGE_SEARCH_MODE.reject()
    mode, purge = await get_args(event.message)
    for i in image_urls:
        msg_list = await image_search(i, mode, purge, PROXY)
        if isinstance(event, PrivateMessageEvent):
            for msg in msg_list:
                await bot.send_private_msg(user_id=event.user_id, message=msg)
        elif isinstance(event, GroupMessageEvent):
            if config.group_forward_search_result and len(msg_list) > 1:
                await bot.send_group_forward_msg(
                    group_id=event.group_id,
                    messages=[
                        {
                            "type": "node",
                            "data": {
                                "name": "\u200b",
                                "uin": bot.self_id,
                                "content": msg,
                            },
                        }
                        for msg in msg_list
                    ],
                )
            else:
                for msg in msg_list:
                    await bot.send_group_msg(group_id=event.group_id, message=msg)



import base64
import re
from io import BytesIO

import httpx
from pyquery import PyQuery


# 将图片转化为 base64
async def get_pic_base64_by_url(url: str, proxy: str) -> str:
    async with httpx.AsyncClient(proxies=proxy) as client:
        r = await client.get(url)
        image_buffer = BytesIO(r.content)
        return str(base64.b64encode(image_buffer.getvalue()), encoding="utf-8")


async def handle_img(url: str, proxy: str, hide_img: bool) -> str:
    if hide_img:
        return ""
    img_base64 = await get_pic_base64_by_url(url, proxy)
    if img_base64:
        return f"[CQ:image,file=base64://{img_base64}]"
    return f"图片下载失败: {url}"


async def get_source(url: str, proxy: str) -> str:
    source = ""
    async with httpx.AsyncClient(proxies=proxy) as client:
        if "danbooru.donmai.us" in url:
            source = PyQuery((await client.get(url)).content)(".image-container").attr(
                "data-normalized-source"
            )
        elif "konachan.com" in url or "yande.re" in url:
            source = PyQuery((await client.get(url)).content)(
                "#stats li:contains(Source) a"
            ).attr("href")
        elif "gelbooru.com" in url:
            source = PyQuery((await client.get(url)).content)(
                "#tag-list li:contains(Source) a"
            ).attr("href")
    if source:
        return str(source)
    return ""


async def shorten_pixiv_url(url: str) -> str:
    pid_search = re.compile(r"pixiv.+(?:illust_id=|artworks/)([0-9]+)")
    uid_search = re.compile(r"pixiv.+(?:member\.php\?id=|users/)([0-9]+)")
    if pid_search.search(url):
        return f"https://pixiv.net/i/{pid_search.search(url).group(1)}"
    if uid_search.search(url):
        return f"https://pixiv.net/u/{uid_search.search(url).group(1)}"
    return url



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['YetAnotherPicSearch']

package_data = \
{'': ['*']}

install_requires = \
['PicImageSearch>=2.2.6,<3.0.0',
 'arrow>=1.2.0,<2.0.0',
 'httpx>=0.18.0,<0.19.0',
 'nonebot-adapter-onebot>=2.0.0b1,<3.0.0',
 'nonebot2>=2.0.0b1,<3.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'pyquery>=1.4.3,<2.0.0',
 'tinydb>=4.6.1,<5.0.0']

setup_kwargs = {
    'name': 'yetanotherpicsearch',
    'version': '1.2.2',
    'description': 'Yet Another Picture Search Nonebot Plugin',
    'long_description': '# YetAnotherPicSearch\n\n基于 [nonebot/nonebot2](https://github.com/nonebot/nonebot2) 及 [kitUIN/PicImageSearch](https://github.com/kitUIN/PicImageSearch) 的另一个 Nonebot 搜图插件。\n\n目前适配的是 `OneBot V11` ，没适配 QQ 频道。\n\n主要受到 [Tsuk1ko/cq-picsearcher-bot](https://github.com/Tsuk1ko/cq-picsearcher-bot) 的启发，以及其遇到的问题 [Tsuk1ko/cq-picsearcher-bot#283](https://github.com/Tsuk1ko/cq-picsearcher-bot/issues/283) 不好解决。\n\n此外，我只需要基础的搜图功能，加上一些小调整，于是忍不住自己也写了一个。\n',
    'author': 'NekoAria',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NekoAria/YetAnotherPicSearch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
